#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
A class for some standard image quality metrics
"""
import numpy as np

class QualityTools:
    def __init__(self, im1, im2):
        if im1.size != im2.size:
            print ('Error: Sizes of images/volumes are different')
            raise SystemExit
        self.im1 = im1 # image or volume - 1
        self.im2 = im2 # image or volume - 2
    def nrmse(self):
        """ Normalised Root Mean Square Error """
        rmse = np.sqrt(np.sum((self.im2 - self.im1) ** 2) / float(self.im1.size))
        max_val = max(np.max(self.im1), np.max(self.im2))
        min_val = min(np.min(self.im1), np.min(self.im2))
        return 1 - (rmse / (max_val - min_val))
    def rmse(self):
        """ Root Mean Square Error """
        rmse = np.sqrt(np.sum((self.im1 - self.im2) ** 2) / float(self.im1.size))
        return rmse
    def ssim(self, window, k=(0.01, 0.03), l=255):
        from scipy.signal import fftconvolve
        """See https://ece.uwaterloo.ca/~z70wang/research/ssim/"""
        # Check if the window is smaller than the images.
        for a, b in zip(window.shape, self.im1.shape):
            if a > b:
                return None, None
        # Values in k must be positive according to the base implementation.
        for ki in k:
            if ki < 0:
                return None, None
    
        c1 = (k[0] * l) ** 2
        c2 = (k[1] * l) ** 2
        window = window/np.sum(window)
    
        mu1 = fftconvolve(self.im1, window, mode='valid')
        mu2 = fftconvolve(self.im2, window, mode='valid')
        mu1_sq = mu1 * mu1
        mu2_sq = mu2 * mu2
        mu1_mu2 = mu1 * mu2
        sigma1_sq = fftconvolve(self.im1 * self.im1, window, mode='valid') - mu1_sq
        sigma2_sq = fftconvolve(self.im2 * self.im2, window, mode='valid') - mu2_sq
        sigma12 = fftconvolve(self.im1 * self.im2, window, mode='valid') - mu1_mu2
    
        if c1 > 0 and c2 > 0:
            num = (2 * mu1_mu2 + c1) * (2 * sigma12 + c2)
            den = (mu1_sq + mu2_sq + c1) * (sigma1_sq + sigma2_sq + c2)
            ssim_map = num / den
        else:
            num1 = 2 * mu1_mu2 + c1
            num2 = 2 * sigma12 + c2
            den1 = mu1_sq + mu2_sq + c1
            den2 = sigma1_sq + sigma2_sq + c2
            ssim_map = np.ones(np.shape(mu1))
            index = (den1 * den2) > 0
            ssim_map[index] = (num1[index] * num2[index]) / (den1[index] * den2[index])
            index = (den1 != 0) & (den2 == 0)
            ssim_map[index] = num1[index] / den1[index]    
        mssim = ssim_map.mean()
        return mssim, ssim_map



import unittest
import numpy as np
import os
import timeit
from ccpi.filters.regularisers import (
    ROF_TV,
    FGP_TV,
    PD_TV,
    SB_TV,
    TGV,
    LLT_ROF,
    FGP_dTV,
    NDF,
    Diff4th,
)
from ccpi.filters.utils import cilregcuda

gpu_modules_available = cilregcuda is not None
from testroutines import BinReader, rmse, printParametersToString


@unittest.skipUnless(gpu_modules_available, "Skipping as GPU modules not available")
class TestRegularisers(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(__file__), "test_imageLena.bin")
        # lena_gray_512.tif

    def _initiate_data(self):
        plt = BinReader()
        # read image
        Im = plt.imread(self.filename)
        Im = np.asarray(Im, dtype="float32")

        Im = Im / 255
        perc = 0.05
        slices = 20

        noisyVol = np.zeros((slices, 512, 512), dtype="float32")
        noisyRef = np.zeros((slices, 512, 512), dtype="float32")
        idealVol = np.zeros((slices, 512, 512), dtype="float32")

        for i in range(slices):
            noisyVol[i, :, :] = Im + np.random.normal(
                loc=0, scale=perc * Im, size=np.shape(Im)
            )
            noisyRef[i, :, :] = Im + np.random.normal(
                loc=0, scale=0.01 * Im, size=np.shape(Im)
            )
            idealVol[i, :, :] = Im
        return noisyVol, noisyRef, idealVol

    def test_ROF_TV_CPU_vs_GPU(self):

        noisyVol, noisyRef, idealVol = self._initiate_data()

        print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
        print("____________ROF-TV bench___________________")
        print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")

        # set parameters
        pars = {
            "algorithm": ROF_TV,
            "input": noisyVol,
            "regularisation_parameter": 0.02,
            "number_of_iterations": 100,
            "time_marching_parameter": 0.001,
            "tolerance_constant": 0.0,
        }
        print("#############ROF TV CPU####################")
        start_time = timeit.default_timer()
        rof_cpu = ROF_TV(
            pars["input"],
            pars["regularisation_parameter"],
            pars["number_of_iterations"],
            pars["time_marching_parameter"],
            pars["tolerance_constant"],
            device="cpu",
        )
        rms = rmse(idealVol, rof_cpu)
        pars["rmse"] = rms

        txtstr = printParametersToString(pars)
        txtstr += "%s = %.3fs" % ("elapsed time", timeit.default_timer() - start_time)
        print(txtstr)
        print("##############ROF TV GPU##################")
        start_time = timeit.default_timer()
        rof_gpu = ROF_TV(
            pars["input"],
            pars["regularisation_parameter"],
            pars["number_of_iterations"],
            pars["time_marching_parameter"],
            pars["tolerance_constant"],
            device="gpu",
        )

        rms = rmse(idealVol, rof_gpu)
        pars["rmse"] = rms
        pars["algorithm"] = ROF_TV
        txtstr = printParametersToString(pars)
        txtstr += "%s = %.3fs" % ("elapsed time", timeit.default_timer() - start_time)
        print(txtstr)
        print("--------Compare the results--------")
        tolerance = 1e-05
        diff_im = np.zeros(np.shape(rof_cpu))
        diff_im = abs(rof_cpu - rof_gpu)
        diff_im[diff_im > tolerance] = 1
        self.assertLessEqual(diff_im.sum(), 1)

    def test_FGP_TV_CPU_vs_GPU(self):
        noisyVol, noisyRef, idealVol = self._initiate_data()

        print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")
        print("____________FGP-TV bench___________________")
        print("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%")

        # set parameters
        pars = {
            "algorithm": FGP_TV,
            "input": noisyVol,
            "regularisation_parameter": 0.05,
            "number_of_iterations": 200,
            "tolerance_constant": 0.0,
            "methodTV": 0,
            "nonneg": 0,
        }

        print("#############FGP TV CPU####################")
        start_time = timeit.default_timer()
        fgp_cpu = FGP_TV(
            pars["input"],
            pars["regularisation_parameter"],
            pars["number_of_iterations"],
            pars["tolerance_constant"],
            pars["methodTV"],
            pars["nonneg"],
            device="cpu",
        )

        rms = rmse(idealVol, fgp_cpu)
        pars["rmse"] = rms

        txtstr = printParametersToString(pars)
        txtstr += "%s = %.3fs" % ("elapsed time", timeit.default_timer() - start_time)
        print(txtstr)

        print("##############FGP TV GPU##################")
        start_time = timeit.default_timer()
        fgp_gpu = FGP_TV(
            pars["input"],
            pars["regularisation_parameter"],
            pars["number_of_iterations"],
            pars["tolerance_constant"],
            pars["methodTV"],
            pars["nonneg"],
            device="gpu",
        )

        rms = rmse(idealVol, fgp_gpu)
        pars["rmse"] = rms
        pars["algorithm"] = FGP_TV
        txtstr = printParametersToString(pars)
        txtstr += "%s = %.3fs" % ("elapsed time", timeit.default_timer() - start_time)
        print(txtstr)

        print("--------Compare the results--------")
        tolerance = 1e-05
        diff_im = abs(fgp_cpu - fgp_gpu)
        np.testing.assert_array_less(diff_im, tolerance)

if __name__ == "__main__":
    unittest.main()



import platform
import ctypes
import os
import warnings

try:
    pre = {'Linux': 'lib', 'Windows': '', 'Darwin': 'lib'}[platform.system()]
except KeyError:
    raise ValueError(f"unsupported platform: {platform.system()}")
else:
    ext = {'Linux': '.so', 'Windows': '.dll', 'Darwin': '.dylib'}[platform.system()]

_here = os.path.dirname(__file__)
dll = f'{pre}cilreg{ext}'
cilreg = ctypes.cdll.LoadLibrary(os.path.join(_here, dll))

gpudll = f'{pre}cilregcuda{ext}'
try:
    cilregcuda = ctypes.cdll.LoadLibrary(os.path.join(_here, gpudll))
except Exception as exc:
    warnings.warn(str(exc), ImportWarning, stacklevel=2)
    warnings.warn(f"Found: {os.listdir(_here)}", ImportWarning, stacklevel=2)
    cilregcuda = None
